###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def list(*, user_id=_ELIDE, screen_name=_ELIDE, reverse=_ELIDE):
    """
    Returns all lists the authenticating or specified user subscribes to,
    including their own. The user is specified using the

    :param user_id: The ID of the user for whom to return results. Helpful for
        disambiguating when a valid user ID is also a valid screen name. Note:
        : Specifies the ID of the user to get lists from. Helpful for
        disambiguating when a valid user ID is also a valid screen name.

    :param screen_name: The screen name of the user for whom to return results.
        Helpful for disambiguating when a valid screen name is also a user ID.

    :param reverse: Set this to true if you would like owned lists to be
        returned first. See description above for information on how this
        parameter works.
    """
    binding = {'user_id': user_id, 'screen_name': screen_name, 'reverse':
               reverse}
    url = 'https://api.twitter.com/1.1/lists/list.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:lists',
                           'get-lists-list',
                           binding)


def members(list_id, slug, *, owner_screen_name=_ELIDE, owner_id=_ELIDE,
            count=_ELIDE, cursor=_ELIDE, include_entities=_ELIDE,
            skip_status=_ELIDE):
    """
    Returns the members of the specified list. Private list members will only
    be shown if the authenticated user owns the specified list.

    :param list_id: The numerical id of the list.

    :param slug: You can identify a list by its slug instead of its numerical
        id. If you decide to do so, note that you’ll also have to specify the
        list owner using the owner_id or owner_screen_name parameters.

    :param owner_screen_name: The screen name of the user who owns the list
        being requested by a slug.

    :param owner_id: The user ID of the user who owns the list being requested
        by a slug.

    :param count: Specifies the number of results to return per page (see
        cursor below). The default is 20, with a maximum of 5,000.

    :param cursor: Causes the collection of list members to be broken into
        “pages” of consistent sizes (specified by the count parameter). If no
        cursor is provided, a value of -1 will be assumed, which is the first
        “page.” The response from the API will include a previous_cursor and
        next_cursor to allow paging back and forth. See Using cursors to
        navigate collections for more information.

    :param include_entities: The entities node will not be included when set to
        false.

    :param skip_status: When set to either true, t or 1 statuses will not be
        included in the returned user objects.
    """
    binding = {'list_id': list_id, 'slug': slug, 'owner_screen_name':
               owner_screen_name, 'owner_id': owner_id, 'count': count,
               'cursor': cursor, 'include_entities': include_entities,
               'skip_status': skip_status}
    url = 'https://api.twitter.com/1.1/lists/members.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:lists',
                           'get-lists-members',
                           binding)


def members_show(list_id, slug, user_id, screen_name, *,
                 owner_screen_name=_ELIDE, owner_id=_ELIDE,
                 include_entities=_ELIDE, skip_status=_ELIDE):
    """
    Check if the specified user is a member of the specified list.

    :param list_id: The numerical id of the list.

    :param slug: You can identify a list by its slug instead of its numerical
        id. If you decide to do so, note that you’ll also have to specify the
        list owner using the owner_id or owner_screen_name parameters.

    :param user_id: The ID of the user for whom to return results for. Helpful
        for disambiguating when a valid user ID is also a valid screen name.

    :param screen_name: The screen name of the user for whom to return results
        for. Helpful for disambiguating when a valid screen name is also a user
        ID.

    :param owner_screen_name: The screen name of the user who owns the list
        being requested by a slug.

    :param owner_id: The user ID of the user who owns the list being requested
        by a slug.

    :param include_entities: When set to either true, t or 1, each tweet will
        include a node called “entities”. This node offers a variety of
        metadata about the tweet in a discreet structure, including:
        user_mentions, urls, and hashtags. While entities are opt-in on
        timelines at present, they will be made a default component of output
        in the future. See Tweet Entities for more details.

    :param skip_status: When set to either true, t or 1 statuses will not be
        included in the returned user objects.
    """
    binding = {'list_id': list_id, 'slug': slug, 'user_id': user_id,
               'screen_name': screen_name, 'owner_screen_name':
               owner_screen_name, 'owner_id': owner_id, 'include_entities':
               include_entities, 'skip_status': skip_status}
    url = 'https://api.twitter.com/1.1/lists/members/show.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:lists',
                           'get-lists-members-show',
                           binding)


def memberships(*, user_id=_ELIDE, screen_name=_ELIDE, count=_ELIDE,
                cursor=_ELIDE, filter_to_owned_lists=_ELIDE):
    """
    Returns the lists the specified user has been added to. If

    :param user_id: The ID of the user for whom to return results for. Helpful
        for disambiguating when a valid user ID is also a valid screen name.

    :param screen_name: The screen name of the user for whom to return results
        for. Helpful for disambiguating when a valid screen name is also a user
        ID.

    :param count: The amount of results to return per page. Defaults to 20. No
        more than 1000 results will ever be returned in a single page.

    :param cursor: Breaks the results into pages. Provide a value of -1 to
        begin paging. Provide values as returned in the response body’s
        next_cursor and previous_cursor attributes to page back and forth in
        the list. It is recommended to always use cursors when the method
        supports them. See [node:10362] for more information.

    :param filter_to_owned_lists: When set to true, t or 1, will return just
        lists the authenticating user owns, and the user represented by user_id
        or screen_name is a member of.
    """
    binding = {'user_id': user_id, 'screen_name': screen_name, 'count': count,
               'cursor': cursor, 'filter_to_owned_lists':
               filter_to_owned_lists}
    url = 'https://api.twitter.com/1.1/lists/memberships.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:lists',
                           'get-lists-memberships',
                           binding)


def ownerships(*, user_id=_ELIDE, screen_name=_ELIDE, count=_ELIDE,
               cursor=_ELIDE):
    """
    Returns the lists owned by the specified Twitter user. Private lists will
    only be shown if the authenticated user is also the owner of the lists.

    :param user_id: The ID of the user for whom to return results for.

    :param screen_name: The screen name of the user for whom to return results
        for.

    :param count: The amount of results to return per page. Defaults to 20. No
        more than 1000 results will ever be returned in a single page.

    :param cursor: Breaks the results into pages. Provide a value of -1 to
        begin paging. Provide values as returned in the response body’s
        next_cursor and previous_cursor attributes to page back and forth in
        the list. It is recommended to always use cursors when the method
        supports them. See [node:10362] for more information.
    """
    binding = {'user_id': user_id, 'screen_name': screen_name, 'count': count,
               'cursor': cursor}
    url = 'https://api.twitter.com/1.1/lists/ownerships.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:lists',
                           'get-lists-ownerships',
                           binding)


def show(list_id, slug, *, owner_screen_name=_ELIDE, owner_id=_ELIDE):
    """
    Returns the specified list. Private lists will only be shown if the
    authenticated user owns the specified list.

    :param list_id: The numerical id of the list.

    :param slug: You can identify a list by its slug instead of its numerical
        id. If you decide to do so, note that you’ll also have to specify the
        list owner using the owner_id or owner_screen_name parameters.

    :param owner_screen_name: The screen name of the user who owns the list
        being requested by a slug.

    :param owner_id: The user ID of the user who owns the list being requested
        by a slug.
    """
    binding = {'list_id': list_id, 'slug': slug, 'owner_screen_name':
               owner_screen_name, 'owner_id': owner_id}
    url = 'https://api.twitter.com/1.1/lists/show.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:lists',
                           'get-lists-show',
                           binding)


def statuses(list_id, slug, *, owner_screen_name=_ELIDE, owner_id=_ELIDE,
             since_id=_ELIDE, max_id=_ELIDE, count=_ELIDE,
             include_entities=_ELIDE, include_rts=_ELIDE):
    """
    Returns a timeline of tweets authored by members of the specified list.
    Retweets are included by default. Use the

    :param list_id: The numerical id of the list.

    :param slug: You can identify a list by its slug instead of its numerical
        id. If you decide to do so, note that you’ll also have to specify the
        list owner using the owner_id or owner_screen_name parameters.

    :param owner_screen_name: The screen name of the user who owns the list
        being requested by a slug.

    :param owner_id: The user ID of the user who owns the list being requested
        by a slug.

    :param since_id: Returns results with an ID greater than (that is, more
        recent than) the specified ID. There are limits to the number of Tweets
        which can be accessed through the API. If the limit of Tweets has
        occured since the since_id, the since_id will be forced to the oldest
        ID available.

    :param max_id: Returns results with an ID less than (that is, older than)
        or equal to the specified ID.

    :param count: Specifies the number of results to retrieve per “page.”

    :param include_entities: Entities are ON by default in API 1.1, each tweet
        includes a node called “entities”. This node offers a variety of
        metadata about the tweet in a discreet structure, including:
        user_mentions, urls, and hashtags. You can omit entities from the
        result by using include_entities=false

    :param include_rts: When set to either true, t or 1, the list timeline will
        contain native retweets (if they exist) in addition to the standard
        stream of tweets. The output format of retweeted tweets is identical to
        the representation you see in home_timeline.
    """
    binding = {'list_id': list_id, 'slug': slug, 'owner_screen_name':
               owner_screen_name, 'owner_id': owner_id, 'since_id': since_id,
               'max_id': max_id, 'count': count, 'include_entities':
               include_entities, 'include_rts': include_rts}
    url = 'https://api.twitter.com/1.1/lists/statuses.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:lists',
                           'get-lists-statuses',
                           binding)


def subscribers(list_id, slug, *, owner_screen_name=_ELIDE, owner_id=_ELIDE,
                count=_ELIDE, cursor=_ELIDE, include_entities=_ELIDE,
                skip_status=_ELIDE):
    """
    Returns the subscribers of the specified list. Private list subscribers
    will only be shown if the authenticated user owns the specified list.

    :param list_id: The numerical id of the list.

    :param slug: You can identify a list by its slug instead of its numerical
        id. If you decide to do so, note that you’ll also have to specify the
        list owner using the owner_id or owner_screen_name parameters.

    :param owner_screen_name: The screen name of the user who owns the list
        being requested by a slug.

    :param owner_id: The user ID of the user who owns the list being requested
        by a slug.

    :param count: Specifies the number of results to return per page (see
        cursor below). The default is 20, with a maximum of 5,000.

    :param cursor: Breaks the results into pages. A single page contains 20
        lists. Provide a value of -1 to begin paging. Provide values as
        returned in the response body’s next_cursor and previous_cursor
        attributes to page back and forth in the list. See Using cursors to
        navigate collections for more information.

    :param include_entities: When set to either true, t or 1, each tweet will
        include a node called “entities”. This node offers a variety of
        metadata about the tweet in a discreet structure, including:
        user_mentions, urls, and hashtags. While entities are opt-in on
        timelines at present, they will be made a default component of output
        in the future. See Tweet Entities for more details.

    :param skip_status: When set to either true, t or 1 statuses will not be
        included in the returned user objects.
    """
    binding = {'list_id': list_id, 'slug': slug, 'owner_screen_name':
               owner_screen_name, 'owner_id': owner_id, 'count': count,
               'cursor': cursor, 'include_entities': include_entities,
               'skip_status': skip_status}
    url = 'https://api.twitter.com/1.1/lists/subscribers.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:lists',
                           'get-lists-subscribers',
                           binding)


def subscribers_show(list_id, slug, user_id, screen_name, *,
                     owner_screen_name=_ELIDE, owner_id=_ELIDE,
                     include_entities=_ELIDE, skip_status=_ELIDE):
    """
    Check if the specified user is a subscriber of the specified list. Returns
    the user if they are subscriber.

    :param owner_screen_name: The screen name of the user who owns the list
        being requested by a slug.

    :param owner_id: The user ID of the user who owns the list being requested
        by a slug.

    :param list_id: The numerical id of the list.

    :param slug: You can identify a list by its slug instead of its numerical
        id. If you decide to do so, note that you’ll also have to specify the
        list owner using the owner_id or owner_screen_name parameters.

    :param user_id: The ID of the user for whom to return results for. Helpful
        for disambiguating when a valid user ID is also a valid screen name.

    :param screen_name: The screen name of the user for whom to return results
        for. Helpful for disambiguating when a valid screen name is also a user
        ID.

    :param include_entities: When set to either true, t or 1, each tweet will
        include a node called “entities”. This node offers a variety of
        metadata about the tweet in a discreet structure, including:
        user_mentions, urls, and hashtags. While entities are opt-in on
        timelines at present, they will be made a default component of output
        in the future. See Tweet Entities for more details.

    :param skip_status: When set to either true, t or 1 statuses will not be
        included in the returned user objects.
    """
    binding = {'owner_screen_name': owner_screen_name, 'owner_id': owner_id,
               'list_id': list_id, 'slug': slug, 'user_id': user_id,
               'screen_name': screen_name, 'include_entities':
               include_entities, 'skip_status': skip_status}
    url = 'https://api.twitter.com/1.1/lists/subscribers/show.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:lists',
                           'get-lists-subscribers-show',
                           binding)


def subscriptions(*, user_id=_ELIDE, screen_name=_ELIDE, count=_ELIDE,
                  cursor=_ELIDE):
    """
    Obtain a collection of the lists the specified user is subscribed to, 20
    lists per page by default. Does not include the user’s own lists.

    :param user_id: The ID of the user for whom to return results for. Helpful
        for disambiguating when a valid user ID is also a valid screen name.

    :param screen_name: The screen name of the user for whom to return results
        for. Helpful for disambiguating when a valid screen name is also a user
        ID.

    :param count: The amount of results to return per page. Defaults to 20. No
        more than 1000 results will ever be returned in a single page.

    :param cursor: Breaks the results into pages. Provide a value of -1 to
        begin paging. Provide values as returned in the response body’s
        next_cursor and previous_cursor attributes to page back and forth in
        the list. It is recommended to always use cursors when the method
        supports them. See [node:10362] for more information.
    """
    binding = {'user_id': user_id, 'screen_name': screen_name, 'count': count,
               'cursor': cursor}
    url = 'https://api.twitter.com/1.1/lists/subscriptions.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:lists',
                           'get-lists-subscriptions',
                           binding)


def create(name, *, mode=_ELIDE, description=_ELIDE):
    """
    Creates a new list for the authenticated user. Note that you can create up
    to 1000 lists per account.

    :param name: The name for the list. A list’s name must start with a letter
        and can consist only of 25 or fewer letters, numbers, “-”, or “_”
        characters.

    :param mode: Whether your list is public or private. Values can be public
        or private. If no mode is specified the list will be public.

    :param description: The description to give the list.
    """
    binding = {'name': name, 'mode': mode, 'description': description}
    url = 'https://api.twitter.com/1.1/lists/create.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:lists',
                           'post-lists-create',
                           binding)


def destroy(list_id, slug, *, owner_screen_name=_ELIDE, owner_id=_ELIDE):
    """
    Deletes the specified list. The authenticated user must own the list to be
    able to destroy it.

    :param owner_screen_name: The screen name of the user who owns the list
        being requested by a slug.

    :param owner_id: The user ID of the user who owns the list being requested
        by a slug.

    :param list_id: The numerical id of the list.

    :param slug: You can identify a list by its slug instead of its numerical
        id. If you decide to do so, note that you’ll also have to specify the
        list owner using the owner_id or owner_screen_name parameters.
    """
    binding = {'owner_screen_name': owner_screen_name, 'owner_id': owner_id,
               'list_id': list_id, 'slug': slug}
    url = 'https://api.twitter.com/1.1/lists/destroy.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:lists',
                           'post-lists-destroy',
                           binding)


def members_create(list_id, slug, user_id, screen_name, *,
                   owner_screen_name=_ELIDE, owner_id=_ELIDE):
    """
    Add a member to a list. The authenticated user must own the list to be able
    to add members to it. Note that lists cannot have more than 5,000 members.

    :param list_id: The numerical id of the list.

    :param slug: You can identify a list by its slug instead of its numerical
        id. If you decide to do so, note that you’ll also have to specify the
        list owner using the owner_id or owner_screen_name parameters.

    :param user_id: The ID of the user for whom to return results for. Helpful
        for disambiguating when a valid user ID is also a valid screen name.

    :param screen_name: The screen name of the user for whom to return results
        for. Helpful for disambiguating when a valid screen name is also a user
        ID.

    :param owner_screen_name: The screen name of the user who owns the list
        being requested by a slug.

    :param owner_id: The user ID of the user who owns the list being requested
        by a slug.
    """
    binding = {'list_id': list_id, 'slug': slug, 'user_id': user_id,
               'screen_name': screen_name, 'owner_screen_name':
               owner_screen_name, 'owner_id': owner_id}
    url = 'https://api.twitter.com/1.1/lists/members/create.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:lists',
                           'post-lists-members-create',
                           binding)


def members_create_all(list_id, slug, *, user_id=_ELIDE, screen_name=_ELIDE,
                       owner_screen_name=_ELIDE, owner_id=_ELIDE):
    """
    Adds multiple members to a list, by specifying a comma-separated list of
    member ids or screen names. The authenticated user must own the list to be
    able to add members to it. Note that lists can’t have more than 5,000
    members, and you are limited to adding up to 100 members to a list at a
    time with this method.

    :param list_id: The numerical id of the list.

    :param slug: You can identify a list by its slug instead of its numerical
        id. If you decide to do so, note that you’ll also have to specify the
        list owner using the owner_id or owner_screen_name parameters.

    :param user_id: A comma separated list of user IDs, up to 100 are allowed
        in a single request.

    :param screen_name: A comma separated list of screen names, up to 100 are
        allowed in a single request.

    :param owner_screen_name: The screen name of the user who owns the list
        being requested by a slug.

    :param owner_id: The user ID of the user who owns the list being requested
        by a slug.
    """
    binding = {'list_id': list_id, 'slug': slug, 'user_id': user_id,
               'screen_name': screen_name, 'owner_screen_name':
               owner_screen_name, 'owner_id': owner_id}
    url = 'https://api.twitter.com/1.1/lists/members/create_all.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:lists',
                           'post-lists-members-create-all',
                           binding)


def members_destroy(*, list_id=_ELIDE, slug=_ELIDE, user_id=_ELIDE,
                    screen_name=_ELIDE, owner_screen_name=_ELIDE,
                    owner_id=_ELIDE):
    """
    Removes the specified member from the list. The authenticated user must be
    the list’s owner to remove members from the list.

    :param list_id: The numerical id of the list.

    :param slug: You can identify a list by its slug instead of its numerical
        id. If you decide to do so, note that you’ll also have to specify the
        list owner using the owner_id or owner_screen_name parameters.

    :param user_id: The ID of the user to remove from the list. Helpful for
        disambiguating when a valid user ID is also a valid screen name.

    :param screen_name: The screen name of the user for whom to remove from the
        list. Helpful for disambiguating when a valid screen name is also a
        user ID.

    :param owner_screen_name: The screen name of the user who owns the list
        being requested by a slug.

    :param owner_id: The user ID of the user who owns the list being requested
        by a slug.
    """
    binding = {'list_id': list_id, 'slug': slug, 'user_id': user_id,
               'screen_name': screen_name, 'owner_screen_name':
               owner_screen_name, 'owner_id': owner_id}
    url = 'https://api.twitter.com/1.1/lists/members/destroy.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:lists',
                           'post-lists-members-destroy',
                           binding)


def members_destroy_all(list_id, slug, *, user_id=_ELIDE, screen_name=_ELIDE,
                        owner_screen_name=_ELIDE, owner_id=_ELIDE):
    """
    Removes multiple members from a list, by specifying a comma-separated list
    of member ids or screen names. The authenticated user must own the list to
    be able to remove members from it. Note that lists can’t have more than 500
    members, and you are limited to removing up to 100 members to a list at a
    time with this method.

    :param list_id: The numerical id of the list.

    :param slug: You can identify a list by its slug instead of its numerical
        id. If you decide to do so, note that you’ll also have to specify the
        list owner using the owner_id or owner_screen_name parameters.

    :param user_id: A comma separated list of user IDs, up to 100 are allowed
        in a single request.

    :param screen_name: A comma separated list of screen names, up to 100 are
        allowed in a single request.

    :param owner_screen_name: The screen name of the user who owns the list
        being requested by a slug.

    :param owner_id: The user ID of the user who owns the list being requested
        by a slug.
    """
    binding = {'list_id': list_id, 'slug': slug, 'user_id': user_id,
               'screen_name': screen_name, 'owner_screen_name':
               owner_screen_name, 'owner_id': owner_id}
    url = 'https://api.twitter.com/1.1/lists/members/destroy_all.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:lists',
                           'post-lists-members-destroy-all',
                           binding)


def subscribers_create(list_id, slug, *, owner_screen_name=_ELIDE,
                       owner_id=_ELIDE):
    """
    Subscribes the authenticated user to the specified list.

    :param owner_screen_name: The screen name of the user who owns the list
        being requested by a slug.

    :param owner_id: The user ID of the user who owns the list being requested
        by a slug.

    :param list_id: The numerical id of the list.

    :param slug: You can identify a list by its slug instead of its numerical
        id. If you decide to do so, note that you’ll also have to specify the
        list owner using the owner_id or owner_screen_name parameters.
    """
    binding = {'owner_screen_name': owner_screen_name, 'owner_id': owner_id,
               'list_id': list_id, 'slug': slug}
    url = 'https://api.twitter.com/1.1/lists/subscribers/create.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:lists',
                           'post-lists-subscribers-create',
                           binding)


def subscribers_destroy(list_id, slug, *, owner_screen_name=_ELIDE,
                        owner_id=_ELIDE):
    """
    Unsubscribes the authenticated user from the specified list.

    :param list_id: The numerical id of the list.

    :param slug: You can identify a list by its slug instead of its numerical
        id. If you decide to do so, note that you’ll also have to specify the
        list owner using the owner_id or owner_screen_name parameters.

    :param owner_screen_name: The screen name of the user who owns the list
        being requested by a slug.

    :param owner_id: The user ID of the user who owns the list being requested
        by a slug.
    """
    binding = {'list_id': list_id, 'slug': slug, 'owner_screen_name':
               owner_screen_name, 'owner_id': owner_id}
    url = 'https://api.twitter.com/1.1/lists/subscribers/destroy.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:lists',
                           'post-lists-subscribers-destroy',
                           binding)


def update(list_id, slug, *, name=_ELIDE, mode=_ELIDE, description=_ELIDE,
           owner_screen_name=_ELIDE, owner_id=_ELIDE):
    """
    Updates the specified list. The authenticated user must own the list to be
    able to update it.

    :param list_id: The numerical id of the list.

    :param slug: You can identify a list by its slug instead of its numerical
        id. If you decide to do so, note that you’ll also have to specify the
        list owner using the owner_id or owner_screen_name parameters.

    :param name: The name for the list.

    :param mode: Whether your list is public or private. Values can be public
        or private. If no mode is specified the list will be public.

    :param description: The description to give the list.

    :param owner_screen_name: The screen name of the user who owns the list
        being requested by a slug.

    :param owner_id: The user ID of the user who owns the list being requested
        by a slug.
    """
    binding = {'list_id': list_id, 'slug': slug, 'name': name, 'mode': mode,
               'description': description, 'owner_screen_name':
               owner_screen_name, 'owner_id': owner_id}
    url = 'https://api.twitter.com/1.1/lists/update.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:lists',
                           'post-lists-update',
                           binding)


_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/list.json'] = 'https://dev.twitter.com/rest/reference/get/lists/list'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/members.json'] = 'https://dev.twitter.com/rest/reference/get/lists/members'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/members/show.json'] = 'https://dev.twitter.com/rest/reference/get/lists/members/show'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/memberships.json'] = 'https://dev.twitter.com/rest/reference/get/lists/memberships'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/ownerships.json'] = 'https://dev.twitter.com/rest/reference/get/lists/ownerships'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/show.json'] = 'https://dev.twitter.com/rest/reference/get/lists/show'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/statuses.json'] = 'https://dev.twitter.com/rest/reference/get/lists/statuses'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/subscribers.json'] = 'https://dev.twitter.com/rest/reference/get/lists/subscribers'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/subscribers/show.json'] = 'https://dev.twitter.com/rest/reference/get/lists/subscribers/show'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/subscriptions.json'] = 'https://dev.twitter.com/rest/reference/get/lists/subscriptions'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/create.json'] = 'https://dev.twitter.com/rest/reference/post/lists/create'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/destroy.json'] = 'https://dev.twitter.com/rest/reference/post/lists/destroy'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/members/create.json'] = 'https://dev.twitter.com/rest/reference/post/lists/members/create'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/members/create_all.json'] = 'https://dev.twitter.com/rest/reference/post/lists/members/create_all'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/members/destroy.json'] = 'https://dev.twitter.com/rest/reference/post/lists/members/destroy'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/members/destroy_all.json'] = 'https://dev.twitter.com/rest/reference/post/lists/members/destroy_all'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/subscribers/create.json'] = 'https://dev.twitter.com/rest/reference/post/lists/subscribers/create'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/subscribers/destroy.json'] = 'https://dev.twitter.com/rest/reference/post/lists/subscribers/destroy'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/lists/update.json'] = 'https://dev.twitter.com/rest/reference/post/lists/update'



from brittle_wit.rate_limit import TimingRateLimiter


def test_timing_rate_limiter():
    limiter = TimingRateLimiter(60)
    assert limiter.can_proceed()
    assert not limiter.can_proceed()

    limiter._start_time -= 61
    assert limiter.can_proceed()



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def site(follow, *, delimited=_ELIDE, stall_warnings=_ELIDE, with_=_ELIDE,
         replies=_ELIDE, stringify_friend_ids=_ELIDE):
    """
    Streams messages for a set of users, as described in

    :param follow: A comma separated list of user IDs, indicating the users to
        return statuses for in the stream. See follow for more information.

    :param delimited: Specifies whether messages should be length-delimited.
        See delimited for more information.

    :param stall_warnings: Specifies whether stall warnings should be
        delivered. See stall_warnings for more information.

    :param with_: Specifies whether to return information for just the users
        specified in the follow parameter, or include messages from accounts
        they follow. See with

    :param replies: Specifies whether to return additional @replies. See
        replies for more information.

    :param stringify_friend_ids: Specifies whether to send the friends lists
        preamble as an array of integers or an array of strings. See
        stringify_friend_id for more information.
    """
    binding = {'follow': follow, 'delimited': delimited, 'stall_warnings':
               stall_warnings, 'with': with_, 'replies': replies,
               'stringify_friend_ids': stringify_friend_ids}
    url = 'https://sitestream.twitter.com/1.1/site.json'
    return _TwitterRequest('GET',
                           url,
                           'streaming:site',
                           'get-site',
                           binding)


_TwitterRequest.DOC_URLS['https://sitestream.twitter.com/1.1/site.json'] = 'https://dev.twitter.com/streaming/reference/get/site'



import re
from vaquero import transformations as t
from lxml.html import fromstring as _fromstring


def prepare(src, dst):
    # Copy non-private keys to the destination.
    for k in src:
        if not k.startswith('_'):
            dst[k] = src[k]

    # Convert the html to a lxml node
    src['doc'] = _fromstring(src['_raw_html'])
    del src['_raw_html']


MISSING_URLS = {'get-media-upload-status':
                    "https://upload.twitter.com/1.1/media/upload.json",
                'post-media-upload-finalize':
                    "https://upload.twitter.com/1.1/media/upload.json",
                'post-batch-accounts-account-id-line-items':
                    "https://ads-api.twitter.com/1/batch/accounts/:account_id/line_items",
                'post-batch-accounts-account-id-campaigns':
                    "https://ads-api.twitter.com/1/batch/accounts/:account_id/campaigns"}


def _match_first(node, *css_selectors):
    for selector in css_selectors:
        m = node.cssselect(selector)
        if m:
            return m[0]
    return None


def extract_and_verify_basic_info(src, dst):
    doc = src['doc']
    snippet = doc.cssselect("div[itemprop='articleBody'] > div.section")[0]
    dst['service'] = snippet.attrib['id']

    method, *path = snippet.cssselect('h1')[0].text.split(" ")
    assert dst['method'] == method
    dst['path'] = " ".join(path)
    dst['desc'] = snippet.cssselect("p")[0].text

    try:
        url_section = snippet.cssselect("#resource-url")[0]
        m = _match_first(url_section,
                         'span.pre', 'code.docutils', 'p > em', 'p > cite')

        url = (m if m is not None else url_section).text_content().strip()
        dst['url'] = url
    except IndexError:
        patch_url = MISSING_URLS.get(dst['service'])
        if patch_url:
            dst['url'] = patch_url
        else:
            raise


def fix_bad_url(_, dst):
    if "Resource URL¶" in dst['url']:
        dst['url'] = dst['url'].replace("Resource URL¶", "").strip()


RESOURCE_INFO_RENAMES = {'response_formats': 'resp_format',
                         'requires_authentication': 'authentication_required'}


def extract_resource_information(src, dst):
    doc = src['doc']
    snippet = doc.cssselect('div#resource-information tr')
    d = {}
    for row in snippet:
        k, v = [item.text_content() for item in row.cssselect('td')]
        d[k] = v
    t.pythonize_ks(d)
    t.rename_ks(d, RESOURCE_INFO_RENAMES)
    for k, v in d.items():
        if k == 'resp_format':
            src[k] = v
        else:
            dst[k] = v


def rate_limited_to_boolean(src, dst):
    dst['rate_limited'] = dst.get('rate_limited') == 'Yes'


def authentication_required_to_boolean(src, dst):
    b = dst.get('authentication_required', "No") != 'No'
    dst['authentication_required'] = b


FORMAT_MAPPING = {'JSON': 'JSON', "204 - No Content": None}


def extract_resp_formats(src, dst):
    if 'resp_format' in src:
        dst['resp_format'] = FORMAT_MAPPING[src['resp_format']]
    else:
        dst['resp_format'] = None


LIMIT_MAP = {'requests_15_min_window_app_auth': 'app',
             'requests_15_min_window_per_app': 'app',
             'requests_15_min_window_user_auth': 'user'}


def extract_limits(src, dst):
    dst['limits'] = {}

    for k, v in dst.items():
        if not k.startswith('requests'):
            continue

        v = v.strip()

        # Special case.
        if v == '15/user and 750/app':
            dst['limits']['user'] = 15
            dst['limits']['app'] = 750
        elif v == 'Refer to your existing rate limit agreements.':
            service = dst['service']
            assert service == 'post-direct-messages-events-new-message-create'
        else:
            k = LIMIT_MAP[k]
            assert k not in dst['limits']
            dst['limits'][k] = int(v)


def _fix_param_desc(d):
    s = d.get('description')
    if s is None:
        return
    s = s.replace(" . ", ". ").replace(" , ", ", ")
    if s.endswith(" ."):
        s = s[:-2] + "."
    d['description'] = s


def _fix_param_default(d):
    default = d.get('default_value')
    if default is not None and default == '':
        del d['default_value']


TYPE_REGEXP = re.compile(r"([\d\w\_]+)\s+Type: (.*)$")


def _fix_param_type(d):
    m = TYPE_REGEXP.match(d['name'])
    if not m:
        return
    d['name'] = m.group(1)
    d['type'] = m.group(2)


VALID_PARAM_KS = {'name', 'default_value', 'example', 'required',
                  'description', 'constraints'}


def _extract_header_from_rows(rows):
    top_row = rows[0]
    items = [td.text_content() or '' for td in top_row.cssselect('td')]
    if not items:
        items = [td.text_content() or '' for td in top_row.cssselect('th')]
        if not items:
            return None

    t.sstrip_all(items)

    ks = [t.pythonize_identifier(k) for k in items]
    return tuple(ks)


def _tail(items):
    head_consumed = False
    for item in items:
        if not head_consumed:
            head_consumed = True
        else:
            yield item


SPECIAL_NAMES = {'follow see note*': 'follow',
                 'track see note*': 'track',
                 'user_ids DEPRECATED': 'user_ids',
                 'locations see note*': 'locations'}


def _extract_param_rows_normal(header, rows):
    data = []

    for tr in _tail(rows):
        items = [td.text_content() or '' for td in tr.cssselect('td')]
        t.sstrip_all(items)

        d = dict(zip(header, items))

        _fix_param_desc(d)
        _fix_param_default(d)
        _fix_param_type(d)

        # Special cases
        d['name'] = SPECIAL_NAMES.get(d['name'], d['name'])

        data.append(d)

    return data


EMBEDDED_PARAM_RE = re.compile("([\w\d\_]+)\s+\(([^\)]+)\)")


def _extract_param_rows_no_header(rows):
    data = []

    for tr in rows:
        items = [td.text_content() or '' for td in tr.cssselect('td')]
        t.sstrip_all(items)
        k, desc = items

        d = {'desc': desc}
        _fix_param_desc(d)

        name, required = EMBEDDED_PARAM_RE.match(k).groups()
        d['name'] = name
        d['required'] = required
        data.append(d)

    return data


def extract_params(src, dst):
    doc = src['doc']
    rows = doc.cssselect('#parameters > table tr')

    if not rows:
        return

    header, data = _extract_header_from_rows(rows), None

    rows = doc.cssselect('#parameters > table tr')
    if set(header).issubset(VALID_PARAM_KS):
        data = _extract_param_rows_normal(header, rows)
    elif len(header) == 2:
        data = _extract_param_rows_no_header(rows)
    else:
        dst['_bad_ks'] = header

    if data:
        dst['params'] = data


def fix_edge_case_params(src, dst):
    for param in dst.get('params', []):
        parts = param['name'].split(' ')
        if len(parts) == 1:
            continue
        param['name'] = parts[0]
        param['required'] = parts[1]


REQUIRED_VALUES = {'True', 'requied', 'required'}


def normalize_required_params(src, dst):
    for param in dst.get('params', []):
        param['required'] = param.get('required') in REQUIRED_VALUES


def extract_example_request(src, dst):
    doc = src['doc']
    example_req = doc.cssselect('#example-request > p')
    if example_req:
        dst['example_request'] = example_req[0].text_content()


def extract_example_response(src, dst):
    doc = src['doc']
    code_sections = doc.cssselect('#example-response > div.code')
    if code_sections:
        dst['example_response'] = code_sections[0].text_content()


def extract_family(_, dst):
    path = dst['path']

    if path.startswith('/1/'):  # Ads
        path = path.replace("/1/", "")

    sub_family = path.split('/')[0]
    dst['family'] = dst['group'] + ":" + sub_family
    assert sub_family


def fix_path(_, dst):
    if dst['path'].startswith('/'):
        dst['path'] = dst['path'][1:]


def update_for_slugs(_, dst):
    dst['slugs'], path = [], dst['path']

    if ':' not in path:
        return

    url = dst['url']

    if 'params' not in dst:
        dst['params'] = []
    known_params = {param['name'] for param in dst['params']}

    parts = path.split("/")
    slugs = [s for s in parts if s.startswith(':')]

    # Order so none can be a prefix of the next.
    for _, slug in reversed(sorted((len(slug), slug) for slug in slugs)):
        k = slug[1:]
        url = url.replace(slug, "{" + k + "}")
        dst['slugs'].append(k)

        if k not in known_params:
            dst['params'].append({'name': k, 'required': True})

    dst['url'] = url


VALID_GROUPS = {'ads', 'rest', 'streaming', 'webhooks'}

VALID_FAMILES = {'ads:accounts',
                 'ads:batch',
                 'ads:bidding_rules',
                 'ads:conversion_attribution',
                 'ads:conversion_event',
                 'ads:iab_categories',
                 'ads:insights',
                 'ads:line_items',
                 'ads:stats',
                 'ads:tailored_audience_memberships',
                 'ads:targeting_criteria',
                 'rest:account',
                 'rest:application',
                 'rest:blocks',
                 'rest:collections',
                 'rest:direct_messages',
                 'rest:favorites',
                 'rest:followers',
                 'rest:friends',
                 'rest:friendships',
                 'rest:geo',
                 'rest:help',
                 'rest:lists',
                 'rest:media',
                 'rest:mutes',
                 'rest:saved_searches',
                 'rest:search',
                 'rest:statuses',
                 'rest:trends',
                 'rest:users',
                 'streaming:c',
                 'streaming:site',
                 'streaming:statuses',
                 'streaming:user',
                 'webhooks:account_activity'}


VALIDATION_SPEC = {
    'authentication_required': ({True, False}, False),
    'desc': (None, None),
    'family': (VALID_FAMILES, None),
    'group': (VALID_GROUPS, None),
    'method': ({'GET', 'POST', 'DELETE', 'PUT'}, 'GET'),
    'resp_format': ({'JSON', None}, 'JSON'),
    'rate_limited': ({True, False}, True),
    'path': (None, None),
    'url': (None, None),
    'reference_url': (None, None)
}


def validate(_, dst):
    for k, (valid_set, default) in VALIDATION_SPEC.items():
        if k not in dst:
            if default is not None:
                dst[k] = default
            else:
                assert False, "Missing and no default on {}".format(k)
        elif valid_set is not None:
            assert dst[k] in valid_set, dst[k]


PARAM_VALIDATION_SPEC = {
    'name': (None, None),
    'required': ({True, False}, None)
}


def validate_params(_, dst):
    if 'params' not in dst:
        dst['params'] = []
        return

    for param in dst['params']:
        for k, (valid_set, default) in PARAM_VALIDATION_SPEC.items():
            if k not in param:
                if default is not None:
                    param[k] = default
                else:
                    assert False, "Missing and no default on {}".format(k)
            elif valid_set is not None:
                assert param[k] in valid_set, param[k]



import os
import pytest
import shutil
from tests.helpers import FIXTURES_DIR, load_fixture_txt, load_fixture_json
from api_gen.api_code_gen import (pep8_join,
                                  _generate_param_tokens,
                                  generate_source,
                                  _param_name_translations,
                                  _generate_param_doc,
                                  _generate_doc_str,
                                  _generate_def_line,
                                  _generate_binding,
                                  _generate_url_block,
                                  generate_modules,
                                  _generate_func_name)


def test_pep8_join():
    expected = load_fixture_txt("pep8_join.txt").rstrip()
    sent = "this is a sequence of words some are reallyreallyreallylong"
    tokens = sent.split(" ") * 3
    res = pep8_join(tokens, ", ", init_indent='--', next_indent='::')
    assert res == expected


def test_param_name_translations():
    example = ["with", "array[]", "for", "simple"]
    expected = dict(zip(example, ["with_", "array", "for_", "simple"]))
    assert _param_name_translations(example) == expected


DEF_LINE_1 = """
def account_media_by_id_and_account_id_via_fake(id, account_id_, *,
                                                count=_ELIDE, sort_by=_ELIDE,
                                                cursor=_ELIDE,
                                                with_deleted=_ELIDE,
                                                account_media_ids=_ELIDE):
""".strip()


def test_generate_def_line():
    example_api_def = {'family': 'ads:accounts',
                       'group': 'ads',
                       'method': 'FAKE',
                       'params': [{'name': 'count', 'required': False},
                                  {'name': 'account_id', 'required': True},
                                  {'name': 'sort_by', 'required': False},
                                  {'name': 'cursor', 'required': False},
                                  {'name': 'with_deleted', 'required': False},
                                  {'name': 'account_media_ids',
                                      'required': False},
                                  {'name': 'id', 'required': True}],
                       'path': 'accounts/:account_id/account_media/:id',
                       'slugs': ['account_id', 'id'],
                       'service': 'get-accounts-account-id-account-media-id'}
    renamings = {'account_id': 'account_id_'}
    res = _generate_def_line(example_api_def, True, **renamings)
    assert res == DEF_LINE_1

    example_api_def = {'family': 'ads:accounts',
                       'group': 'ads',
                       'method': 'FAKE',
                       'params': [],
                       'path': 'accounts/list_all',
                       'slugs': [],
                       'service': 'get-accounts-account-id-account-media-id'}
    res = _generate_def_line(example_api_def)
    assert res == "def list_all():"


def test_generate_param_doc():
    example_def = {'description': ("This is a description line." +
                                   " It should wrap correctly so it is " +
                                   " PEP8 compliant."),
                   'type': 'An Integer'}
    expected = ("    :param id: This is a description line. It should " +
                "wrap correctly so it is\n" +
                "        PEP8 compliant. (An Integer)")
    assert _generate_param_doc(example_def, 'id') == expected
    assert _generate_param_doc({}, 'id') is None


def test_generate_doc_str():
    example_api_def = {'family': 'ads:accounts',
                       'group': 'ads',
                       'desc': "An Example",
                       'method': 'FAKE',
                       'params': [{'name': 'count', 'desc': 'A word'},
                                  {'name': 'account_id', 'desc': 'A word'},
                                  {'name': 'sort_by', 'desc': 'A word'},
                                  {'name': 'cursor', 'desc': 'A word'},
                                  {'name': 'with_deleted',
                                      'desc': 'A word'},
                                  {'name': 'account_media_ids',
                                      'desc': 'A word'},
                                  {'name': 'id', 'desc': 'A word'}],
                       'path': 'accounts/:account_id/account_media/:id',
                       'slugs': ['account_id', 'id'],
                       'service': 'get-accounts-account-id-account-media-id'}
    renamings = {'account_id': 'account_id_'}
    expected = load_fixture_txt("doc_str.txt")

    res = _generate_doc_str(example_api_def, **renamings)
    assert res == expected


def test_generate_binding():
    names = ["with", "array[]", "for", "simple"]
    renamings = _param_name_translations(names)
    assert _generate_binding({'params': {}}) == "    binding = {}"

    api_def = {'params': [{'name': "with"},
                          {'name': "array[]"},
                          {'name': "for"},
                          {'name': "simple"}]}

    expected = ("    binding = {'with': with_, 'array[]': array, " +
                "'for': for_, 'simple': simple}")
    assert _generate_binding(api_def, **renamings) == expected


def test_generate_url_block():
    expected = "    url = 'http://google.com/{id}'"
    res = _generate_url_block({'url': "http://google.com/{id}",
                               'slugs': []})
    assert res == expected

    res = _generate_url_block({'url': "http://google.com/{id}",
                               'slugs': ['id']})
    assert res == expected + "\n    url = url.format(**binding)"


def test_generate_modules():
    tmp_dir = os.path.join(FIXTURES_DIR, "skel")
    if os.path.exists(tmp_dir):
        shutil.rmtree(tmp_dir)

    definitions = {'public': [{'family': 'x:dog'}, {'family': 'y:cat'}],
                   'private': [{'family': 'y:octopus'}]}

    generate_modules(tmp_dir, definitions, lambda api_def: '')

    expected_paths = [('public_api', '__init__.py'),
                      ('public_api', 'dog.py'),
                      ('public_api', 'cat.py'),
                      ('private_api', 'octopus.py')]

    for path in expected_paths:
        assert os.path.exists(os.path.join(tmp_dir, *path))

    shutil.rmtree(tmp_dir)


def test_generate_signature_str():
    example = {'params': [{'name': 'a', 'required': True},
                          {'name': 'b', 'required': False},
                          {'name': 'c', 'required': False},
                          {'name': 'd', 'required': True}],
               'slugs': ['d']}
    expected = [("d", None), ("a", None), ("*", None),
                ("b", "_ELIDE"), ("c", "_ELIDE")]
    assert _generate_param_tokens(example) == expected


def test_generate_source():
    example = load_fixture_json("get_block_ids.json")
    src = generate_source(example)
    with open(os.path.join(FIXTURES_DIR, "get_block_ids.src"), "w") as fp:
        fp.write(src)
    assert src == load_fixture_txt("get_block_ids.src")


def test_simple():
    example = {'family': 'rest:blocks',
               'method': 'GET',
               'params': [{'name': 'stringify_ids', 'required': False},
                          {'name': 'cursor', 'required': False}],
               'path': 'blocks/ids',
               'slugs': []}

    assert _generate_func_name(example) == "ids"
    assert _generate_func_name(example, True) == "ids_via_get"


def test_long_simple():
    example = {'family': 'rest:direct_messages',
               'method': 'GET',
               'params': [{'name': 'count', 'required': False},
                          {'name': 'cursor', 'required': False}],
               'path': 'direct_messages/events/list',
               'slugs': []}

    assert _generate_func_name(example) == "events_list"
    assert _generate_func_name(example, True) == "events_list_via_get"


def test_longer_simple():
    example = {'family': 'rest:direct_messages',
               'method': 'GET',
               'params': [{'name': 'id', 'required': True}],
               'path': 'direct_messages/welcome_messages/rules/show',
               'slugs': []}

    assert _generate_func_name(example) == "welcome_messages_rules_show"
    assert _generate_func_name(example, True) == \
        "welcome_messages_rules_show_via_get"


def test_slugged():
    example = {'family': 'rest:saved_searches',
               'method': 'GET',
               'params': [{'name': 'id', 'required': True}],
               'path': 'saved_searches/show/:id',
               'slugs': ['id']}

    assert _generate_func_name(example) == "show_by_id"
    assert _generate_func_name(example, True) == "show_by_id_via_get"


def test_long_slugged():
    example = {'family': 'rest:users',
               'method': 'GET',
               'params': [{'name': 'slug', 'required': True}],
               'path': 'users/suggestions/:slug/members',
               'slugs': ['slug']}

    assert _generate_func_name(example) == "suggestions_members_by_slug"
    assert _generate_func_name(example, True) == \
        "suggestions_members_by_slug_via_get"


def test_longer_slugged():
    example = {'family': 'ads:accounts',
               'method': 'GET',
               'params': [{'name': 'count', 'required': False},
                          {'name': 'account_id', 'required': True},
                          {'name': 'sort_by', 'required': False},
                          {'name': 'cursor', 'required': False},
                          {'name': 'with_deleted', 'required': False},
                          {'name': 'account_media_ids', 'required': False},
                          {'name': 'id', 'required': True}],
               'path': 'accounts/:account_id/account_media/:id',
               'slugs': ['account_id', 'id']}

    assert _generate_func_name(example) == "account_media_by_id_and_account_id"
    assert _generate_func_name(example, True) == \
        "account_media_by_id_and_account_id_via_get"


def test_longer_slugged_weird():
    example = {'family': 'ads:stats',
               'method': 'DELETE',
               'params': [{'name': 'account_id', 'required': True},
                          {'name': 'job_id', 'required': True}],
               'path': '1/stats/jobs/accounts/:account_id/:job_id',
               'slugs': ['account_id', 'job_id']}

    assert _generate_func_name(example) == \
        "jobs_accounts_by_job_id_and_account_id"
    assert _generate_func_name(example, True) == \
        "jobs_accounts_by_job_id_and_account_id_via_delete"


def test_too_long():
    example = {'family': 'ads:stats',
               'method': 'DELETE',
               'params': [{'name': 'account_id', 'required': True},
                          {'name': 'another', 'required': True},
                          {'name': 'job_id', 'required': True}],
               'path': '1/stats/jobs/accounts/:account_id/:job_id/:another',
               'slugs': ['account_id', 'job_id', 'another']}

    with pytest.raises(ValueError):
        _generate_func_name(example)


def test_bad_name():
    example = {'family': 'ads:stats',
               'method': 'GET',
               'params': {},
               'path': 'stats',
               'slugs': []}

    assert _generate_func_name(example) == 'stats'

    example['path'] = "stats/arr[]"
    assert _generate_func_name(example) == 'arr'



import asyncio
import aiohttp
import os
from collections import deque
from contextlib import contextmanager
from brittle_wit_core import AppCredentials, ClientCredentials
from brittle_wit.executors import ManagedClientRequestProcessors
from brittle_wit.helpers import running_task_names
import brittle_wit.connection_props as DEFAULTS


class ClientContext:
    """
    Safely-encapsulates ``App``-associated ``ClientCredentials``.
    """
    __slots__ = ('_app', '_client_cred', '__weakref__')

    def __init__(self, app, client_cred):
        """
        :param app: The ``App`` these credentials belong to
        :param client_cred: The client's credentials.
        """
        self._app = app
        self._client_cred = client_cred

    async def execute(self, req, timeout=120, n_retries=5, sleep_time=60):
        """
        Execute the given request in the context of the associated client.

        :param req: a ``TwitterRequest``
        :return: A ``TwitterResponse``
        """
        # Client credentials may change -- they are just oauth key-pairs.
        # XXX: This violates the hash pattern so I thin the API needs a change.
        # I.E "REFRESH".
        # XXX: Figure out weakref pattern.
        processor = self._app.manager[self._client_cred]

        return await processor.execute(self._app.session,
                                       self._app.app_cred,
                                       req,
                                       timeout=timeout,
                                       n_retries=n_retries,
                                       sleep_time=sleep_time)


PRESTINE = 'prestine'
RUNNING = 'running'
FINISHED = 'finished'


class App:
    """
    A simple interface for building a Twitter App.
    """

    @classmethod
    def with_defaults(cls, app_cred, loop=None, session_overrides=None,
                      tcp_conn_overrides=None):
        """
        Create an application with sensible defaults for transports.

        :param app_cred: an `AppCredentials` object
        :param loop: the ``asyncio`` event loop.
        :param session_overrides: overrides for ``aiohttp.ClientSession``
            instantiation.
        :param tcp_conn_overrides``: overrides for ``aiohttp.TCPConnector``
            instantiation.
        """
        tcp_opts = {**DEFAULTS.TCP_CONN_REST, **(tcp_conn_overrides or {})}
        tcp_opts['loop'] = loop
        tcp_connector = aiohttp.TCPConnector(**tcp_opts)

        session_opts = {**DEFAULTS.SESSION_OPTS, **(session_overrides or {})}
        session_opts['connector'] = tcp_connector
        session_opts['loop'] = loop

        session = aiohttp.ClientSession(**session_opts)
        return App(app_cred, session)

    def __init__(self, app_cred, client_session):
        """
        Create a twitter app.

        :param app_cred: an `AppCredentials` object
        :param client_session: a aiohttp ClientSession
        """
        self._app_cred = app_cred
        self._session = client_session

        self._manager = ManagedClientRequestProcessors()
        self._initial_coros = deque()
        self._maintenance_coros = {}
        self._state = PRESTINE

    @property
    def app_cred(self):
        return self._app_cred

    @property
    def session(self):
        return self._session

    @property
    def manager(self):
        return self._manager

    def client_context(self, client_cred):
        return ClientContext(self, client_cred)

    def run(self):
        """
        Run this app (blocking)
        """
        return asyncio.get_event_loop().run_until_complete(self.run_async())

    def schedule(self, coro):
        """
        Add a coroutine to the loop.

        :param coro: the coroutine to add. If the app is running, it is
            immediately added as a task to the event loop. If the app is not
            yet running, it is added to the initial task queue. If the app
            is no longer running, it is an error.
        :return: the task if and only if the app is running, otherwise None
        """
        if self._state == RUNNING:
            return asyncio.get_event_loop().create_task(coro)
        elif self._state == PRESTINE:
            self._initial_coros.append(coro)
        else:
            raise RuntimeError("Scheduled in state {}".format(self._state))

    def schedule_maintenance(self, coro, send_cancel=True):
        """
        Add a coroutine to the loop that does maintenance.

        An app with only maintance functions running may exit.

        :param coro: A coroutine that does maintenance.
        :param send_cancel: if True, the App will cancel this coro
            if it's still running when only maintenance tasks are running
        :return: the task if and only if the app is running, otherwise None
        """
        self._maintenance_coros[coro] = send_cancel

        if self._state == RUNNING:
            return asyncio.get_event_loop().create_task(coro)
        elif self._state == PRESTINE:
            pass  # Only need it in maintenance funcs.
        else:
            raise RuntimeError("Scheduled in state {}".format(self._state))

    def _prepare_for_run(self):
        if self._state == FINISHED:
            raise RuntimeError("This app has already finished! Can't rerun!")
        elif self._state == RUNNING:
            raise RuntimeError("This app is already running!")
        else:
            self._state = RUNNING  # XXX: WAIT WHAT?

    async def run_async(self):
        self._prepare_for_run()

        loop = asyncio.get_event_loop()
        self._maintenance_coros[self._manager.maintain()] = True

        with self._session as session:

            # Add all initial tasks ------------------------------------------

            while self._initial_coros:
                loop.create_task(self._initial_coros.popleft())

            for coro in self._maintenance_coros:
                loop.create_task(coro)

            # Run until only simple maintenance tasks exist ------------------

            while True:
                pending = []

                for task in asyncio.Task.all_tasks():
                    # We can ignore a task if it's already done.
                    if task.done():
                        continue

                    # Obviously, this coro is running.
                    if task._coro.__qualname__ == 'App.run_async':
                        continue

                    # Tasks that are neither done nor a background maintenance
                    # task must be gathered before completion.
                    if task._coro not in self._maintenance_coros:
                        pending.append(task)

                if pending:
                    await asyncio.gather(*pending, return_exceptions=True)
                else:
                    break

        # End all maintenance tasks ------------------------------------------

        tasks_to_cancel = [task for task in asyncio.Task.all_tasks()
                           if not task.done() and
                           task._coro.__qualname__ != 'App.run_async' and
                           self._maintenance_coros.get(task._coro)]

        for task in tasks_to_cancel:
            task.cancel()

        await asyncio.gather(*tasks_to_cancel, return_exceptions=True)

        self._state = 'finished'

    @contextmanager
    def setup_and_run(self):
        yield self
        self.run()



from tests.helpers import load_fixture_txt
import brittle_wit.rest_api as rest_api
from brittle_wit.helpers import _twitter_request_html_repr



def test_twitter_request_html_repr():
    expected = load_fixture_txt("ipython_html.html")
    req = rest_api.direct_messages.new("hello world")
    assert _twitter_request_html_repr(req).strip() == expected.strip()



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def rate_limit_status(*, resources=_ELIDE):
    """
    Returns the current rate limits for methods belonging to the specified
    resource families.

    :param resources: A comma-separated list of resource families you want to
        know the current rate limit disposition for. For best performance, only
        specify the resource families pertinent to your application. See API
        Rate Limiting for more information.
    """
    binding = {'resources': resources}
    url = 'https://api.twitter.com/1.1/application/rate_limit_status.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:application',
                           'get-application-rate-limit-status',
                           binding)


_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/application/rate_limit_status.json'] = 'https://dev.twitter.com/rest/reference/get/application/rate_limit_status'



import pytest
import random
from unittest.mock import MagicMock, patch

from brittle_wit_core import (TwitterRequest, TwitterError, TwitterResponse,
                              AppCredentials, ClientCredentials)
from brittle_wit.rate_limit import RateLimit
from brittle_wit.executors import execute_req
from tests.helpers import MockHTTPReq, MockHTTPResp


pytestmark = pytest.mark.asyncio
client_cred = ClientCredentials(1, "client", "secret")


@pytest.fixture
def twitter_req():
    return TwitterRequest("method", "url", "family", "service")


async def test_good_resp_and_json_body(twitter_req):
    rate_limit = RateLimit.from_ignorance()
    http_req = MockHTTPReq(MockHTTPResp(200, {}, {'count': 100}))

    res = await execute_req(client_cred, twitter_req, http_req, rate_limit)

    assert isinstance(res, TwitterResponse)
    assert res.body['count'] == 100


async def test_good_resp_and_text_body():
    twitter_req.parse_as = 'text'
    rate_limit = RateLimit.from_ignorance()
    http_req = MockHTTPReq(MockHTTPResp(200, {}, "hello world"))

    coro = execute_req(client_cred, twitter_req, http_req, rate_limit)
    res = await coro
    assert isinstance(res, TwitterResponse)
    assert res.body == "hello world"


async def test_bad_resp():
    twitter_req.parse_as = 'text'
    rate_limit = RateLimit.from_ignorance()
    http_req = MockHTTPReq(MockHTTPResp(403, {}, "Forbidden"))

    coro = execute_req(client_cred, twitter_req, http_req, rate_limit)
    with pytest.raises(TwitterError):
        await coro



from aiohttp import Timeout
from brittle_wit_core import TwitterError
from brittle_wit.executors import twitter_req_to_http_req
from brittle_wit.helpers import LOGGER
from brittle_wit.streaming.entry_processor import (EntryProcessor,
                                                   noop_entry_parser)


class TwitterStream:

    def __init__(self, session, app_cred, client_cred, twitter_req,
                 parser=noop_entry_parser, chunk_size=4096,
                 read_hang_timeout=60):
        """
        Initialize a Twitter stream.

        There is no initial connection; Twitter doesn't like idle reader hands.

        :param session: the aiohttp.ClientSession object
        :param app_cred: the AppCredentials
        :param client_cred: the ClientCredentials
        :param twitter_req: the TwitterRequest to initiate streaming
        :param parser: a parser which takes a byte-entry and translates it
            into a message.
        :param chunk_size: the number of bytes to asynchronously read from
            Twitter for each receive
        :param read_hang_timeout: the number of seconds to wait while waiting
            for the *next* read before raising a Timeout
        """
        self._session = session
        self._app_cred = app_cred
        self._client_cred = client_cred
        self._twitter_req = twitter_req
        self._parser = parser
        self._chunk_size = chunk_size
        self._read_hang_timeout = read_hang_timeout

        self._entry_processor = EntryProcessor()
        self._http_req, self._resp = None, None

    def is_open(self):
        """
        :return: True if their is an active connection to Twitter's servers.
        """
        return (self._http_req is not None and self._resp is not None and
                self._resp.connection and not self._resp.connection.closed)

    def _connect(self):
        """
        Connect to Twitter's servers.
        """
        LOGGER.info("Connecting to twitter.")

        # =====================================================================
        # There may only be one set of credentials connected to a streaming
        # endpoint at a time. Preempting an existing connection is
        # error-prone. Complain loudly when a connection already exists.
        # =====================================================================
        if self.is_open():
            LOGGER.critical("Trying to connect an already open stream!")
            raise RuntimeError("Already connected. Call close() first!")

        # =====================================================================
        # Twitter streams should not read timeout since they poll forever.
        # aiohttp implements it such that reads time is cumulative, not
        # episodic, which is wrong for long-polling semantics. Instead, the
        # read_hang_timeout constructor arg controls the amount of time
        # to wait *between* reads.
        # =====================================================================
        self._http_req = twitter_req_to_http_req(self._session,
                                                 self._app_cred,
                                                 self._client_cred,
                                                 self._twitter_req)

    def disconnect(self):
        """
        Disconnect the stream from Twitter's servers.
        """
        LOGGER.info("Disconnecting stream from twitter's server")

        if self._http_req is not None:
            self._http_req.close()  # Force close.
            self._http_req = None

        if self._resp is not None:
            self._resp = None

    def reconnect(self, force_close_if_open=True, clear_prior_messages=False):
        """
        Reconnect to Twitter's servers.

        :param force_close_if_open: if True, close the prior connection, if it
            exists; if false, raise error if it is open
        :param clear_prior_messages: if True, purges all (correct) messages
            waiting for consumption on the internal EntryProcessor from the
            prior connection
        """
        LOGGER.info("Reconnecting stream.")

        # Twitter wants one IP per connection.
        if self.is_open() and force_close_if_open:
            self.disconnect()

        # The initiating request does not change between connection resets,
        # so old messages are consumable.
        if clear_prior_messages:
            self._entry_processor.purge_mailbox()

        # The buffer underling the entry processor should always be in
        # an initially pristine state. A broken connection virtually
        # ensures corruption.
        self._entry_processor.purge_buffer()

        self._connect()

    async def __aiter__(self):
        """
        Commence streaming, returning this object as an iterator.
        """
        LOGGER.info("Creating async iterator.")

        if not self.is_open():
            self.reconnect()  # Ensure an open connection.

        if self._resp is None:
            self._resp = await self._http_req.__aenter__()

            if self._resp.status != 200:
                resp, self._resp = self._resp, None
                LOGGER.error("Response error: {}".format(resp))
                raise TwitterError(self._client_cred,
                                   self._twitter_req,
                                   resp,
                                   await resp.text())

        return self

    async def __anext__(self):
        """
        This reads one message from the stream.
        """
        # If there are no messages ready, read from the stream.
        while not self._entry_processor:
            with Timeout(self._read_hang_timeout):
                chunk = await self._resp.content.read(self._chunk_size)
                if not chunk:
                    LOGGER.error("Read zero bytes on a stream.")
                    break
                self._entry_processor.process(chunk)

        # If there are no messages ready, the stream closed!
        # XXX: TODO: When does this occur. StopAsycIteration is scary.
        if not self._entry_processor:
            LOGGER.error("Stream Closing (non-explicit)")
            raise StopAsyncIteration

        # Take one message.
        return self._parser(self._entry_processor.pop())



import asyncio
import json
import logging
import os
import unittest
from unittest.mock import Mock, patch
from contextlib import contextmanager
from collections import namedtuple
from functools import wraps
from inspect import iscoroutinefunction


SELF_DIR = os.path.dirname(os.path.realpath(__file__))

FIXTURES_DIR = os.path.join(SELF_DIR, "fixtures")

SERIALIZATION_DIR = os.path.join(SELF_DIR, "tmp")


def load_fixture_txt(file_name):
    print(FIXTURES_DIR)
    with open(os.path.join(FIXTURES_DIR, file_name)) as fp:
        return fp.read()


def load_fixture_json(file_name):
    with open(os.path.join(FIXTURES_DIR, file_name)) as fp:
        return json.load(fp)


@contextmanager
def async_patch(*args, **kwargs):
    value = kwargs.get('return_value')
    if 'return_value' in kwargs:
        del kwargs['return_value']

    async def basic_coro():
        return value

    def side_effect(*args, **kwargs):
        return basic_coro()  # Make a new coroutine for each call.

    kwargs['side_effect'] = side_effect

    with unittest.mock.patch(*args, **kwargs) as f:
        yield f


@contextmanager
def async_ignore_sleep():
    with async_patch('asyncio.sleep') as f:
        def sleep_times():
            return [call[0][0] for call in f.call_args_list]
        yield sleep_times


def mock_http_req(status, closed, body_as_text=''):

    async def text():
        return body_as_text

    resp = Mock()
    resp.connection.closed = closed
    resp.status = status
    resp.text = lambda: text()

    class MockReq:
        async def __aenter__(self):
            return resp

    return MockReq()


def ensure_deleted(file_path):
    try:
        os.unlink(file_path)
    except FileNotFoundError:
        pass


def mock_content_reader(messages, delay=0):
    messages = messages[:]

    async def read(n):
        if delay != 0:
            await asyncio.sleep(delay)
        return messages.pop(0) if messages else ''

    resp = Mock()
    resp.content.read = lambda n: read(n)
    return resp


def mock_stream(messages):
    messages = messages[:]

    class MockStream:

        async def __aiter__(self):
            return self

        async def __anext__(self):
            if not messages:
                raise StopAsyncIteration

            msg = messages.pop(0)

            if isinstance(msg, BaseException):
                raise msg
            else:
                return msg

        def disconnect(self):
            pass

    return MockStream()


def mock_resp(headers):
    mock = unittest.mock.MagicMock()
    mock.headers = headers
    return mock


class MockHTTPResp:

    def __init__(self, status, headers, body, delay=0):
        self.status, self.headers, self.body = status, headers, body
        self.delay = delay

    async def json(self):
        if self.delay > 0:
            await asyncio.sleep(self.delay)
        return self.body

    async def text(self):
        if self.delay > 0:
            await asyncio.sleep(self.delay)
        return self.body


class MockHTTPReq:

    def __init__(self, resp):
        self._resp = resp

    async def __aenter__(self):
        return self._resp

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass


def make_coro_runner(f):

    @wraps(f)
    def coro_runner_test(self):
        # Create a new event loop.
        loop = asyncio.new_event_loop()
        loop.set_debug(True)
        logging.getLogger('asyncio').setLevel(logging.WARNING)
        asyncio.set_event_loop(loop)

        # Run it.
        loop.run_until_complete(f(self))

        # Close it.
        loop.close()

    return coro_runner_test


class AsyncSafeTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        updates = {}
        for k, f in vars(cls).items():
            if k.startswith('test') and iscoroutinefunction(f):
                setattr(cls, k, make_coro_runner(f))


# https://blog.miguelgrinberg.com/post/unit-testing-asyncio-code
def AsyncMock(*args, **kwargs):
    m = mock.MagicMock(*args, **kwargs)

    async def mock_coro(*args, **kwargs):
        return m(*args, **kwargs)

    mock_coro.mock = m
    return mock_coro



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def incoming(*, cursor=_ELIDE, stringify_ids=_ELIDE):
    """
    Returns a collection of numeric IDs for every user who has a pending
    request to follow the authenticating user.

    :param cursor: Causes the list of connections to be broken into pages of no
        more than 5000 IDs at a time. The number of IDs returned is not
        guaranteed to be 5000 as suspended users are filtered out after
        connections are queried. If no cursor is provided, a value of -1 will
        be assumed, which is the first “page.” The response from the API will
        include a previous_cursor and next_cursor to allow paging back and
        forth. See [node:10362, title=”Using cursors to navigate collections”]
        for more information.

    :param stringify_ids: Many programming environments will not consume our
        Tweet ids due to their size. Provide this option to have ids returned
        as strings instead. More about [node:194].
    """
    binding = {'cursor': cursor, 'stringify_ids': stringify_ids}
    url = 'https://api.twitter.com/1.1/friendships/incoming.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:friendships',
                           'get-friendships-incoming',
                           binding)


def lookup(*, screen_name=_ELIDE, user_id=_ELIDE):
    """
    Returns the relationships of the authenticating user to the comma-separated
    list of up to 100 screen_names or user_ids provided. Values for

    :param screen_name: A comma separated list of screen names, up to 100 are
        allowed in a single request.

    :param user_id: A comma separated list of user IDs, up to 100 are allowed
        in a single request.
    """
    binding = {'screen_name': screen_name, 'user_id': user_id}
    url = 'https://api.twitter.com/1.1/friendships/lookup.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:friendships',
                           'get-friendships-lookup',
                           binding)


def no_retweets_ids(*, stringify_ids=_ELIDE):
    """
    Returns a collection of user_ids that the currently authenticated user does
    not want to receive retweets from.

    :param stringify_ids: Many programming environments will not consume our
        ids due to their size. Provide this option to have ids returned as
        strings instead. Read more about [node:194]. This parameter is
        especially important to use in Javascript environments.
    """
    binding = {'stringify_ids': stringify_ids}
    url = 'https://api.twitter.com/1.1/friendships/no_retweets/ids.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:friendships',
                           'get-friendships-no-retweets-ids',
                           binding)


def outgoing(*, cursor=_ELIDE, stringify_ids=_ELIDE):
    """
    Returns a collection of numeric IDs for every protected user for whom the
    authenticating user has a pending follow request.

    :param cursor: Causes the list of connections to be broken into pages of no
        more than 5000 IDs at a time. The number of IDs returned is not
        guaranteed to be 5000 as suspended users are filtered out after
        connections are queried. If no cursor is provided, a value of -1 will
        be assumed, which is the first “page.” The response from the API will
        include a previous_cursor and next_cursor to allow paging back and
        forth. See [node:10362, title=”Using cursors to navigate collections”]
        for more information.

    :param stringify_ids: Many programming environments will not consume our
        Tweet ids due to their size. Provide this option to have ids returned
        as strings instead. More about [node:194].
    """
    binding = {'cursor': cursor, 'stringify_ids': stringify_ids}
    url = 'https://api.twitter.com/1.1/friendships/outgoing.format'
    return _TwitterRequest('GET',
                           url,
                           'rest:friendships',
                           'get-friendships-outgoing',
                           binding)


def show(*, source_id=_ELIDE, source_screen_name=_ELIDE, target_id=_ELIDE,
         target_screen_name=_ELIDE):
    """
    Returns detailed information about the relationship between two arbitrary
    users.

    :param source_id: The user_id of the subject user.

    :param source_screen_name: The screen_name of the subject user.

    :param target_id: The user_id of the target user.

    :param target_screen_name: The screen_name of the target user.
    """
    binding = {'source_id': source_id, 'source_screen_name':
               source_screen_name, 'target_id': target_id,
               'target_screen_name': target_screen_name}
    url = 'https://api.twitter.com/1.1/friendships/show.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:friendships',
                           'get-friendships-show',
                           binding)


def create(*, screen_name=_ELIDE, user_id=_ELIDE, follow=_ELIDE):
    """
    Allows the authenticating users to follow the user specified in the ID
    parameter.

    :param screen_name: The screen name of the user for whom to befriend.

    :param user_id: The ID of the user for whom to befriend.

    :param follow: Enable notifications for the target user.
    """
    binding = {'screen_name': screen_name, 'user_id': user_id, 'follow':
               follow}
    url = 'https://api.twitter.com/1.1/friendships/create.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:friendships',
                           'post-friendships-create',
                           binding)


def destroy(*, screen_name=_ELIDE, user_id=_ELIDE):
    """
    Allows the authenticating user to unfollow the user specified in the ID
    parameter.

    :param screen_name: The screen name of the user for whom to unfollow.

    :param user_id: The ID of the user for whom to unfollow.
    """
    binding = {'screen_name': screen_name, 'user_id': user_id}
    url = 'https://api.twitter.com/1.1/friendships/destroy.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:friendships',
                           'post-friendships-destroy',
                           binding)


def update(*, screen_name=_ELIDE, user_id=_ELIDE, device=_ELIDE,
           retweets=_ELIDE):
    """
    Allows one to enable or disable retweets and device notifications from the
    specified user.

    :param screen_name: The screen name of the user for whom to befriend.

    :param user_id: The ID of the user for whom to befriend.

    :param device: Enable/disable device notifications from the target user.

    :param retweets: Enable/disable retweets from the target user.
    """
    binding = {'screen_name': screen_name, 'user_id': user_id, 'device':
               device, 'retweets': retweets}
    url = 'https://api.twitter.com/1.1/friendships/update.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:friendships',
                           'post-friendships-update',
                           binding)


_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/friendships/incoming.json'] = 'https://dev.twitter.com/rest/reference/get/friendships/incoming'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/friendships/lookup.json'] = 'https://dev.twitter.com/rest/reference/get/friendships/lookup'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/friendships/no_retweets/ids.json'] = 'https://dev.twitter.com/rest/reference/get/friendships/no_retweets/ids'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/friendships/outgoing.format'] = 'https://dev.twitter.com/rest/reference/get/friendships/outgoing'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/friendships/show.json'] = 'https://dev.twitter.com/rest/reference/get/friendships/show'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/friendships/create.json'] = 'https://dev.twitter.com/rest/reference/post/friendships/create'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/friendships/destroy.json'] = 'https://dev.twitter.com/rest/reference/post/friendships/destroy'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/friendships/update.json'] = 'https://dev.twitter.com/rest/reference/post/friendships/update'



import aiohttp


# ============================================================================

TCP_CONN_REST = {}

# I think this is nessessary. If you are streaming and don't
# properly disconnect, Twitter may think you have two IPs open.
# > Some ssl servers do not properly complete ssl shutdown process,
# > in that case asyncio leaks ssl connections. If this parameter is
# > set to True, aiohttp additionally aborts underlining transport
# > after 2 seconds. It is off by default.
TCP_CONN_REST['enable_cleanup_closed'] = True

# ============================================================================

TCP_CONN_STREAMING = TCP_CONN_REST.copy()

TCP_CONN_STREAMING['enable_cleanup_closed'] = True

# From: https://dev.twitter.com/streaming/public
#
# > Each account may create only one standing connection to the public
# > endpoints, and connecting to a public stream more than once with the
# > same account credentials will cause the oldest connection to be
# > disconnected.  Clients which make excessive connection attempts
# > (both successful and unsuccessful) run the risk of having their IP
# > automatically banned.
#
# So, this forces close to ensure only ONE standing connection.
TCP_CONN_STREAMING['force_close'] = True

# TODO: Does this mean you can only stream from one APP or CLIENT per IP?
#
# I think the latter.
#
# As per: https://dev.twitter.com/streaming/userstreams
# > Minimize the number of connections your application makes to User
# > Streams. Each Twitter account is limited to only a few simultaneous
# > User Streams connections per OAuth application, regardless of IP.
# TCP_CONN_STREAMING['limit_per_host'] = 1

# ============================================================================

SESSION_OPTS = {}

# Twitter does not need any cookie processing for it's API.
SESSION_OPTS['cookie_jar'] = aiohttp.DummyCookieJar()

# From: https://dev.twitter.com/streaming/overview/connecting
# > Set a timer, either a 90 second TCP level socket timeout, or a
# > 90 second application level timer on the receipt of new data. If
# > 90 seconds pass with no data received, including newlines,
# > disconnect and reconnect immediately according to the backoff
# > strategies in the next section. The Streaming API will send a
# > keep-alive newline every 30 seconds to prevent your application
# > from timing out the connection. You should wait at least 3 cycles
# > to prevent spurious reconnects in the event of network congestion,
# > local CPU starvation, local GC pauses, etc.
SESSION_OPTS['conn_timeout'] = 90
SESSION_OPTS['read_timeout'] = None



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def ids(*, user_id=_ELIDE, screen_name=_ELIDE, cursor=_ELIDE,
        stringify_ids=_ELIDE, count=_ELIDE):
    """
    Returns a cursored collection of user IDs for every user following the
    specified user.

    :param user_id: The ID of the user for whom to return results for.

    :param screen_name: The screen name of the user for whom to return results
        for.

    :param cursor: Causes the list of connections to be broken into pages of no
        more than 5000 IDs at a time. The number of IDs returned is not
        guaranteed to be 5000 as suspended users are filtered out after
        connections are queried. If no cursor is provided, a value of -1 will
        be assumed, which is the first “page.” The response from the API will
        include a previous_cursor and next_cursor to allow paging back and
        forth. See Using cursors to navigate collections for more information.

    :param stringify_ids: Many programming environments will not consume our
        Tweet ids due to their size. Provide this option to have ids returned
        as strings instead. More about [node:194].

    :param count: Specifies the number of IDs attempt retrieval of, up to a
        maximum of 5,000 per distinct request. The value of count is best
        thought of as a limit to the number of results to return. When using
        the count parameter with this method, it is wise to use a consistent
        count value across all requests to the same user’s collection. Usage of
        this parameter is encouraged in environments where all 5,000 IDs
        constitutes too large of a response.
    """
    binding = {'user_id': user_id, 'screen_name': screen_name, 'cursor':
               cursor, 'stringify_ids': stringify_ids, 'count': count}
    url = 'https://api.twitter.com/1.1/followers/ids.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:followers',
                           'get-followers-ids',
                           binding)


def list(*, user_id=_ELIDE, screen_name=_ELIDE, cursor=_ELIDE, count=_ELIDE,
         skip_status=_ELIDE, include_user_entities=_ELIDE):
    """
    Returns a cursored collection of user objects for users following the
    specified user.

    :param user_id: The ID of the user for whom to return results for.

    :param screen_name: The screen name of the user for whom to return results
        for.

    :param cursor: Causes the results to be broken into pages. If no cursor is
        provided, a value of -1 will be assumed, which is the first “page.” The
        response from the API will include a previous_cursor and next_cursor to
        allow paging back and forth. See Using cursors to navigate collections
        for more information.

    :param count: The number of users to return per page, up to a maximum of
        200. Defaults to 20.

    :param skip_status: When set to either true, t or 1, statuses will not be
        included in the returned user objects. If set to any other value,
        statuses will be included.

    :param include_user_entities: The user object entities node will not be
        included when set to false.
    """
    binding = {'user_id': user_id, 'screen_name': screen_name, 'cursor':
               cursor, 'count': count, 'skip_status': skip_status,
               'include_user_entities': include_user_entities}
    url = 'https://api.twitter.com/1.1/followers/list.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:followers',
                           'get-followers-list',
                           binding)


_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/followers/ids.json'] = 'https://dev.twitter.com/rest/reference/get/followers/ids'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/followers/list.json'] = 'https://dev.twitter.com/rest/reference/get/followers/list'
